@@ -29,6 +29,9 @@
 
 BSON_BEGIN_DECLS
 
+#define mcommon_thread_create COMMON_NAME (thread_create)
+#define mcommon_thread_join COMMON_NAME (thread_join)
+
 #if defined(BSON_OS_UNIX)
 #include <pthread.h>
 
@@ -107,20 +110,27 @@ typedef struct {
 #define BSON_THREAD_FUN(_function_name, _arg_name) \
    unsigned (__stdcall _function_name) (void *(_arg_name))
 #define BSON_THREAD_FUN_TYPE(_function_name) \
-   unsigned(__stdcall * _function_name) (void *)
+   unsigned (__stdcall * _function_name) (void *)
 #define BSON_THREAD_RETURN return 0
 #endif
 
 /* Functions that require definitions get the common prefix (_mongoc for
  * libmongoc or _bson for libbson) to avoid duplicate symbols when linking both
  * libbson and libmongoc statically. */
-int COMMON_PREFIX (thread_join) (bson_thread_t thread);
-int COMMON_PREFIX (thread_create) (bson_thread_t *thread,
-                                   BSON_THREAD_FUN_TYPE (func),
-                                   void *arg);
+int
+mcommon_thread_join (bson_thread_t thread);
+// mcommon_thread_create returns 0 on success. Returns a non-zero error code on
+// error. Callers may use `bson_strerror_r` to get an error message from the
+// returned error code.
+int
+mcommon_thread_create (bson_thread_t *thread,
+                       BSON_THREAD_FUN_TYPE (func),
+                       void *arg);
 
 #if defined(MONGOC_ENABLE_DEBUG_ASSERTIONS) && defined(BSON_OS_UNIX)
-bool COMMON_PREFIX (mutex_is_locked) (bson_mutex_t *mutex);
+#define mcommon_mutex_is_locked COMMON_NAME (mutex_is_locked)
+bool
+mcommon_mutex_is_locked (bson_mutex_t *mutex);
 #endif
 
 /**
