@@ -370,8 +370,9 @@ _noop (void)
                                  read_state_names[bson->read_state]);          \
       return;                                                                  \
    }
-#define HANDLE_OPTION(_key, _type, _state)                                     \
-   (len == strlen (_key) && strncmp ((const char *) val, (_key), len) == 0)    \
+#define HANDLE_OPTION(_selection_statement, _key, _type, _state)               \
+   _selection_statement (len == strlen (_key) &&                               \
+                         strncmp ((const char *) val, (_key), len) == 0)       \
    {                                                                           \
       if (bson->bson_type && bson->bson_type != (_type)) {                     \
          _bson_json_read_set_error (reader,                                    \
@@ -387,15 +388,17 @@ _noop (void)
    }
 
 
-
 bson_json_opts_t *
 bson_json_opts_new (bson_json_mode_t mode, int32_t max_len)
 {
    bson_json_opts_t *opts;
 
    opts = (bson_json_opts_t *) bson_malloc (sizeof *opts);
-   opts->mode = mode;
-   opts->max_len = max_len;
+   *opts = (bson_json_opts_t){
+      .mode = mode,
+      .max_len = max_len,
+      .is_outermost_array = false,
+   };
 
    return opts;
 }
@@ -628,7 +631,13 @@ _bson_json_read_integer (bson_json_reader_t *reader, uint64_t val, int64_t sign)
          bson_append_int32 (
             STACK_BSON_CHILD, key, (int) len, (int) (val * sign));
       } else if (sign == -1) {
+#if defined(_WIN32) && !defined(__MINGW32__)
+         // Unary negation of unsigned integer is deliberate.
+#pragma warning(suppress : 4146)
          bson_append_int64 (STACK_BSON_CHILD, key, (int) len, (int64_t) -val);
+#else
+         bson_append_int64 (STACK_BSON_CHILD, key, (int) len, (int64_t) -val);
+#endif // defined(_WIN32) && !defined(__MINGW32__)
       } else {
          bson_append_int64 (STACK_BSON_CHILD, key, (int) len, (int64_t) val);
       }
@@ -735,33 +744,19 @@ _bson_json_parse_double (bson_json_reader_t *reader,
    *d = strtod (val, NULL);
 
 #ifdef _MSC_VER
+   const double pos_inf = INFINITY;
+   const double neg_inf = -pos_inf;
+
    /* Microsoft's strtod parses "NaN", "Infinity", "-Infinity" as 0 */
    if (*d == 0.0) {
       if (!_strnicmp (val, "nan", vlen)) {
-#ifdef NAN
          *d = NAN;
-#else
-         /* Visual Studio 2010 doesn't define NAN or INFINITY
-          * https://msdn.microsoft.com/en-us/library/w22adx1s(v=vs.100).aspx */
-         unsigned long nan[2] = {0xffffffff, 0x7fffffff};
-         *d = *(double *) nan;
-#endif
          return true;
       } else if (!_strnicmp (val, "infinity", vlen)) {
-#ifdef INFINITY
-         *d = INFINITY;
-#else
-         unsigned long inf[2] = {0x00000000, 0x7ff00000};
-         *d = *(double *) inf;
-#endif
+         *d = pos_inf;
          return true;
       } else if (!_strnicmp (val, "-infinity", vlen)) {
-#ifdef INFINITY
-         *d = -INFINITY;
-#else
-         unsigned long inf[2] = {0x00000000, 0xfff00000};
-         *d = *(double *) inf;
-#endif
+         *d = neg_inf;
          return true;
       }
    }
@@ -831,7 +826,7 @@ static bool
 _unhexlify_uuid (const char *uuid, uint8_t *out, size_t max)
 {
    unsigned int byte;
-   int x = 0;
+   size_t x = 0;
    int i = 0;
 
    BSON_ASSERT (strlen (uuid) == 32);
@@ -866,7 +861,7 @@ _bson_json_parse_binary_elem (bson_json_reader_t *reader,
 
    if (bs == BSON_JSON_LF_BINARY) {
       data->binary.has_binary = true;
-      binary_len = COMMON_PREFIX (bson_b64_pton (val_w_null, NULL, 0));
+      binary_len = mcommon_b64_pton (val_w_null, NULL, 0);
       if (binary_len < 0) {
          _bson_json_read_set_error (
             reader,
@@ -875,9 +870,9 @@ _bson_json_parse_binary_elem (bson_json_reader_t *reader,
       }
 
       _bson_json_buf_ensure (&bson->bson_type_buf[0], (size_t) binary_len + 1);
-      if (COMMON_PREFIX (bson_b64_pton (val_w_null,
-                                        bson->bson_type_buf[0].buf,
-                                        (size_t) binary_len + 1) < 0)) {
+      if (mcommon_b64_pton (val_w_null,
+                            bson->bson_type_buf[0].buf,
+                            (size_t) binary_len + 1) < 0) {
          _bson_json_read_set_error (
             reader,
             "Invalid input string \"%s\", looking for base64-encoded binary",
@@ -970,6 +965,34 @@ _bson_json_parse_binary_elem (bson_json_reader_t *reader,
    }
 }
 
+static bool
+_bson_json_allow_embedded_nulls (bson_json_reader_t const *reader)
+{
+   const bson_json_read_state_t read_state = reader->bson.read_state;
+   const bson_json_read_bson_state_t bson_state = reader->bson.bson_state;
+
+   if (read_state == BSON_JSON_IN_BSON_TYPE_REGEX_VALUES) {
+      if (bson_state == BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN ||
+          bson_state == BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS) {
+         /* Prohibit embedded NULL bytes for canonical extended regex:
+          * { $regularExpression: { pattern: "pattern", options: "options" } }
+          */
+         return false;
+      }
+   }
+
+   if (read_state == BSON_JSON_IN_BSON_TYPE) {
+      if (bson_state == BSON_JSON_LF_REGEX ||
+          bson_state == BSON_JSON_LF_OPTIONS) {
+         /* Prohibit embedded NULL bytes for legacy regex:
+          * { $regex: "pattern", $options: "options" } */
+         return false;
+      }
+   }
+
+   /* Embedded nulls are okay in any other context */
+   return true;
+}
 
 static void
 _bson_json_read_string (bson_json_reader_t *reader, /* IN */
@@ -978,13 +1001,14 @@ _bson_json_read_string (bson_json_reader_t *reader, /* IN */
 {
    bson_json_read_state_t rs;
    bson_json_read_bson_state_t bs;
+   const bool allow_null = _bson_json_allow_embedded_nulls (reader);
 
    BASIC_CB_PREAMBLE;
 
    rs = bson->read_state;
    bs = bson->bson_state;
 
-   if (!bson_utf8_validate ((const char *) val, vlen, true /*allow null*/)) {
+   if (!bson_utf8_validate ((const char *) val, vlen, allow_null)) {
       _bson_json_read_corrupt (reader, "invalid bytes in UTF8 string");
       return;
    }
@@ -1100,9 +1124,9 @@ _bson_json_read_string (bson_json_reader_t *reader, /* IN */
       } break;
       case BSON_JSON_LF_DECIMAL128: {
          bson_decimal128_t decimal128;
-         bson_decimal128_from_string (val_w_null, &decimal128);
 
-         if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
+         if (bson_decimal128_from_string (val_w_null, &decimal128) &&
+             bson->read_state == BSON_JSON_IN_BSON_TYPE) {
             bson->bson_type_data.v_decimal128.value = decimal128;
          } else {
             goto BAD_PARSE;
@@ -1145,18 +1169,60 @@ _bson_json_read_start_map (bson_json_reader_t *reader) /* IN */
    BASIC_CB_PREAMBLE;
 
    if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
-      if (bson->bson_state == BSON_JSON_LF_DATE) {
+      switch (bson->bson_state) {
+      case BSON_JSON_LF_DATE:
          bson->read_state = BSON_JSON_IN_BSON_TYPE_DATE_NUMBERLONG;
-      } else if (bson->bson_state == BSON_JSON_LF_BINARY) {
+         break;
+      case BSON_JSON_LF_BINARY:
          bson->read_state = BSON_JSON_IN_BSON_TYPE_BINARY_VALUES;
-      } else if (bson->bson_state == BSON_JSON_LF_TYPE) {
+         break;
+      case BSON_JSON_LF_TYPE:
          /* special case, we started parsing {$type: {$numberInt: "2"}} and we
           * expected a legacy Binary format. now we see the second "{", so
           * backtrack and parse $type query operator. */
          bson->read_state = BSON_JSON_IN_START_MAP;
+         BSON_ASSERT (bson_in_range_unsigned (int, len));
          STACK_PUSH_DOC (bson_append_document_begin (
-            STACK_BSON_PARENT, key, len, STACK_BSON_CHILD));
+            STACK_BSON_PARENT, key, (int) len, STACK_BSON_CHILD));
          _bson_json_save_map_key (bson, (const uint8_t *) "$type", 5);
+         break;
+      case BSON_JSON_LF_CODE:
+      case BSON_JSON_LF_DECIMAL128:
+      case BSON_JSON_LF_DOUBLE:
+      case BSON_JSON_LF_INT32:
+      case BSON_JSON_LF_INT64:
+      case BSON_JSON_LF_MAXKEY:
+      case BSON_JSON_LF_MINKEY:
+      case BSON_JSON_LF_OID:
+      case BSON_JSON_LF_OPTIONS:
+      case BSON_JSON_LF_REGEX:
+      /**
+       * NOTE: A read_state of BSON_JSON_IN_BSON_TYPE is used when "$regex" is
+       * found, but BSON_JSON_IN_BSON_TYPE_REGEX_STARTMAP is used for
+       * "$regularExpression", which will instead go to a below 'if else' branch
+       * instead of this switch statement. They're both called "regex" in their
+       * respective enumerators, but they behave differently when parsing.
+       */
+      // fallthrough
+      case BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS:
+      case BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN:
+      case BSON_JSON_LF_SYMBOL:
+      case BSON_JSON_LF_UNDEFINED:
+      case BSON_JSON_LF_UUID:
+         // These special keys do not expect objects as their values. Fail.
+         _bson_json_read_set_error (
+            reader,
+            "Unexpected nested object value for \"%s\" key",
+            reader->bson.unescaped.buf);
+         break;
+      case BSON_JSON_LF_DBPOINTER:
+      case BSON_JSON_LF_SCOPE:
+      case BSON_JSON_LF_TIMESTAMP_I:
+      case BSON_JSON_LF_TIMESTAMP_T:
+      default:
+         // These special LF keys aren't handled with BSON_JSON_IN_BSON_TYPE
+         BSON_UNREACHABLE (
+            "These LF values are handled with a different read_state");
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_TIMESTAMP_STARTMAP) {
       bson->read_state = BSON_JSON_IN_BSON_TYPE_TIMESTAMP_VALUES;
@@ -1231,7 +1297,7 @@ _bson_json_read_code_or_scope_key (bson_json_reader_bson_t *bson,
          /* save the key, e.g. {"key": {"$code": "return x", "$scope":{"x":1}}},
           * in case it is overwritten while parsing scope sub-object */
          _bson_json_buf_set (
-            &bson->code_data.key_buf, bson->key_buf.buf, bson->key_buf.len);
+            &bson->code_data.key_buf, bson->key, bson->key_buf.len);
       }
 
       if (is_scope) {
@@ -1269,7 +1335,7 @@ _bson_json_read_map_key (bson_json_reader_t *reader, /* IN */
 {
    bson_json_reader_bson_t *bson = &reader->bson;
 
-   if (!bson_utf8_validate ((const char *) val, len, true /* allow null */)) {
+   if (!bson_utf8_validate ((const char *) val, len, false /* allow null */)) {
       _bson_json_read_corrupt (reader, "invalid bytes in UTF8 string");
       return;
    }
@@ -1300,90 +1366,92 @@ _bson_json_read_map_key (bson_json_reader_t *reader, /* IN */
    }
 
    if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
-      if
-         HANDLE_OPTION ("$regex", BSON_TYPE_REGEX, BSON_JSON_LF_REGEX)
-      else if
-         HANDLE_OPTION ("$options", BSON_TYPE_REGEX, BSON_JSON_LF_OPTIONS)
-      else if
-         HANDLE_OPTION ("$oid", BSON_TYPE_OID, BSON_JSON_LF_OID)
-      else if
-         HANDLE_OPTION ("$binary", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
-      else if
-         HANDLE_OPTION ("$type", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
-      else if
-         HANDLE_OPTION ("$uuid", BSON_TYPE_BINARY, BSON_JSON_LF_UUID)
-      else if
-         HANDLE_OPTION ("$date", BSON_TYPE_DATE_TIME, BSON_JSON_LF_DATE)
-      else if
-         HANDLE_OPTION (
-            "$undefined", BSON_TYPE_UNDEFINED, BSON_JSON_LF_UNDEFINED)
-      else if
-         HANDLE_OPTION ("$minKey", BSON_TYPE_MINKEY, BSON_JSON_LF_MINKEY)
-      else if
-         HANDLE_OPTION ("$maxKey", BSON_TYPE_MAXKEY, BSON_JSON_LF_MAXKEY)
-      else if
-         HANDLE_OPTION ("$numberInt", BSON_TYPE_INT32, BSON_JSON_LF_INT32)
-      else if
-         HANDLE_OPTION ("$numberLong", BSON_TYPE_INT64, BSON_JSON_LF_INT64)
-      else if
-         HANDLE_OPTION ("$numberDouble", BSON_TYPE_DOUBLE, BSON_JSON_LF_DOUBLE)
-      else if
-         HANDLE_OPTION ("$symbol", BSON_TYPE_SYMBOL, BSON_JSON_LF_SYMBOL)
-      else if
-         HANDLE_OPTION (
-            "$numberDecimal", BSON_TYPE_DECIMAL128, BSON_JSON_LF_DECIMAL128)
-      else if (!strcmp ("$timestamp", (const char *) val)) {
+      HANDLE_OPTION (if, "$regex", BSON_TYPE_REGEX, BSON_JSON_LF_REGEX)
+      HANDLE_OPTION (else if, "$options", BSON_TYPE_REGEX, BSON_JSON_LF_OPTIONS)
+      HANDLE_OPTION (else if, "$oid", BSON_TYPE_OID, BSON_JSON_LF_OID)
+      HANDLE_OPTION (else if, "$binary", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
+      HANDLE_OPTION (else if, "$type", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
+      HANDLE_OPTION (else if, "$uuid", BSON_TYPE_BINARY, BSON_JSON_LF_UUID)
+      HANDLE_OPTION (else if, "$date", BSON_TYPE_DATE_TIME, BSON_JSON_LF_DATE)
+      HANDLE_OPTION (
+         else if, "$undefined", BSON_TYPE_UNDEFINED, BSON_JSON_LF_UNDEFINED)
+      HANDLE_OPTION (else if, "$minKey", BSON_TYPE_MINKEY, BSON_JSON_LF_MINKEY)
+      HANDLE_OPTION (else if, "$maxKey", BSON_TYPE_MAXKEY, BSON_JSON_LF_MAXKEY)
+      HANDLE_OPTION (else if, "$numberInt", BSON_TYPE_INT32, BSON_JSON_LF_INT32)
+      HANDLE_OPTION (
+         else if, "$numberLong", BSON_TYPE_INT64, BSON_JSON_LF_INT64)
+      HANDLE_OPTION (
+         else if, "$numberDouble", BSON_TYPE_DOUBLE, BSON_JSON_LF_DOUBLE)
+      HANDLE_OPTION (else if, "$symbol", BSON_TYPE_SYMBOL, BSON_JSON_LF_SYMBOL)
+      HANDLE_OPTION (else if,
+                     "$numberDecimal",
+                     BSON_TYPE_DECIMAL128,
+                     BSON_JSON_LF_DECIMAL128)
+      else if (!strcmp ("$timestamp", (const char *) val))
+      {
          bson->bson_type = BSON_TYPE_TIMESTAMP;
          bson->read_state = BSON_JSON_IN_BSON_TYPE_TIMESTAMP_STARTMAP;
-      } else if (!strcmp ("$regularExpression", (const char *) val)) {
+      }
+      else if (!strcmp ("$regularExpression", (const char *) val))
+      {
          bson->bson_type = BSON_TYPE_REGEX;
          bson->read_state = BSON_JSON_IN_BSON_TYPE_REGEX_STARTMAP;
-      } else if (!strcmp ("$dbPointer", (const char *) val)) {
+      }
+      else if (!strcmp ("$dbPointer", (const char *) val))
+      {
          /* start parsing "key": {"$dbPointer": {...}}, save "key" for later */
          _bson_json_buf_set (
-            &bson->dbpointer_key, bson->key_buf.buf, bson->key_buf.len);
+            &bson->dbpointer_key, bson->key, bson->key_buf.len);
 
          bson->bson_type = BSON_TYPE_DBPOINTER;
          bson->read_state = BSON_JSON_IN_BSON_TYPE_DBPOINTER_STARTMAP;
-      } else if (!strcmp ("$code", (const char *) val)) {
+      }
+      else if (!strcmp ("$code", (const char *) val))
+      {
          _bson_json_read_code_or_scope_key (
             bson, false /* is_scope */, val, len);
-      } else if (!strcmp ("$scope", (const char *) val)) {
+      }
+      else if (!strcmp ("$scope", (const char *) val))
+      {
          _bson_json_read_code_or_scope_key (
             bson, true /* is_scope */, val, len);
-      } else {
+      }
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_DATE_NUMBERLONG) {
-      if
-         HANDLE_OPTION ("$numberLong", BSON_TYPE_DATE_TIME, BSON_JSON_LF_INT64)
-      else {
+      HANDLE_OPTION (if, "$numberLong", BSON_TYPE_DATE_TIME, BSON_JSON_LF_INT64)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_TIMESTAMP_VALUES) {
-      if
-         HANDLE_OPTION ("t", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_T)
-      else if
-         HANDLE_OPTION ("i", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_I)
-      else {
+      HANDLE_OPTION (if, "t", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_T)
+      HANDLE_OPTION (
+         else if, "i", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_I)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_REGEX_VALUES) {
-      if
-         HANDLE_OPTION (
-            "pattern", BSON_TYPE_REGEX, BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN)
-      else if
-         HANDLE_OPTION (
-            "options", BSON_TYPE_REGEX, BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS)
-      else {
+      HANDLE_OPTION (if,
+                     "pattern",
+                     BSON_TYPE_REGEX,
+                     BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN)
+      HANDLE_OPTION (else if,
+                     "options",
+                     BSON_TYPE_REGEX,
+                     BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_BINARY_VALUES) {
-      if
-         HANDLE_OPTION ("base64", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
-      else if
-         HANDLE_OPTION ("subType", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
-      else {
+      HANDLE_OPTION (if, "base64", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
+      HANDLE_OPTION (else if, "subType", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else {
@@ -1942,6 +2010,9 @@ _push_callback (jsonsl_t json,
 {
    bson_json_reader_t *reader = (bson_json_reader_t *) json->data;
 
+   BSON_UNUSED (action);
+   BSON_UNUSED (buf);
+
    switch (state->type) {
    case JSONSL_T_STRING:
    case JSONSL_T_HKEY:
@@ -1973,6 +2044,8 @@ _pop_callback (jsonsl_t json,
    double d;
    const char *obj_text;
 
+   BSON_UNUSED (action);
+
    reader = (bson_json_reader_t *) json->data;
    reader_bson = &reader->bson;
 
@@ -2040,6 +2113,8 @@ _error_callback (jsonsl_t json,
 {
    bson_json_reader_t *reader = (bson_json_reader_t *) json->data;
 
+   BSON_UNUSED (state);
+
    if (err == JSONSL_ERROR_CANT_INSERT && *errat == '{') {
       /* start the next document */
       reader->should_reset = true;
@@ -2154,8 +2229,11 @@ bson_json_reader_read (bson_json_reader_t *reader, /* IN */
 
          /* accumulate a key or string value */
          if (reader->json_text_pos != -1) {
-            if (reader->json_text_pos < reader->json->pos) {
-               accum = BSON_MIN (reader->json->pos - reader->json_text_pos, r);
+            if (bson_cmp_less_su (reader->json_text_pos, reader->json->pos)) {
+               BSON_ASSERT (
+                  bson_in_range_unsigned (ssize_t, reader->json->pos));
+               accum = BSON_MIN (
+                  (ssize_t) reader->json->pos - reader->json_text_pos, r);
                /* if this chunk stopped mid-token, buf_offset is how far into
                 * our current chunk the token begins. */
                buf_offset = AT_LEAST_0 (reader->json_text_pos - start_pos);
@@ -2190,7 +2268,9 @@ bson_json_reader_new (void *data,               /* IN */
    bson_json_reader_t *r;
    bson_json_reader_producer_t *p;
 
-   r = bson_malloc0 (sizeof *r);
+   BSON_UNUSED (allow_multiple);
+
+   r = BSON_ALIGNED_ALLOC0 (bson_json_reader_t);
    r->json = jsonsl_new (STACK_MAX);
    r->json->error_callback = _error_callback;
    r->json->action_callback_PUSH = _push_callback;
@@ -2258,6 +2338,14 @@ bson_json_reader_destroy (bson_json_reader_t *reader) /* IN */
 }
 
 
+void
+bson_json_opts_set_outermost_array (bson_json_opts_t *opts,
+                                    bool is_outermost_array)
+{
+   opts->is_outermost_array = is_outermost_array;
+}
+
+
 typedef struct {
    const uint8_t *data;
    size_t len;
