@@ -21,6 +21,7 @@
 
 #include <bson/bson.h>
 
+#include "mcd-rpc.h"
 #include "mongoc-array-private.h"
 #include "mongoc-buffer-private.h"
 #include "mongoc-config.h"
@@ -44,19 +45,14 @@ BSON_BEGIN_DECLS
 typedef struct _mongoc_cluster_node_t {
    mongoc_stream_t *stream;
    char *connection_address;
-   uint32_t generation;
-
-   /* TODO CDRIVER-3653, these fields are unused. */
-   int32_t max_wire_version;
-   int32_t min_wire_version;
-   int32_t max_write_batch_size;
-   int32_t max_bson_obj_size;
-   int32_t max_msg_size;
+   /* handshake_sd is a server description created from the handshake on the
+    * stream. */
+   mongoc_server_description_t *handshake_sd;
 } mongoc_cluster_node_t;
 
 typedef struct _mongoc_cluster_t {
    int64_t operation_id;
-   uint32_t request_id;
+   int32_t request_id;
    uint32_t sockettimeoutms;
    uint32_t socketcheckintervalms;
    mongoc_uri_t *uri;
@@ -100,17 +96,29 @@ mongoc_cluster_check_interval (mongoc_cluster_t *cluster, uint32_t server_id);
 bool
 mongoc_cluster_legacy_rpc_sendv_to_server (
    mongoc_cluster_t *cluster,
-   mongoc_rpc_t *rpcs,
+   mcd_rpc_message *rpc,
    mongoc_server_stream_t *server_stream,
    bson_error_t *error);
 
 bool
 mongoc_cluster_try_recv (mongoc_cluster_t *cluster,
-                         mongoc_rpc_t *rpc,
+                         mcd_rpc_message *rpc,
                          mongoc_buffer_t *buffer,
                          mongoc_server_stream_t *server_stream,
                          bson_error_t *error);
 
+/**
+ * @brief Obtain a server stream appropriate for read operations on the
+ * cluster.
+ *
+ * Returns a new stream (that must be freed) or NULL and sets an error via
+ * `error`.
+ *
+ * @note The returned stream must be released via
+ * `mongoc_server_stream_cleanup`.
+ *
+ * @note May add nodes and/or update the cluster's topology.
+ */
 mongoc_server_stream_t *
 mongoc_cluster_stream_for_reads (mongoc_cluster_t *cluster,
                                  const mongoc_read_prefs_t *read_prefs,
@@ -118,12 +126,58 @@ mongoc_cluster_stream_for_reads (mongoc_cluster_t *cluster,
                                  bson_t *reply,
                                  bson_error_t *error);
 
+/**
+ * @brief Obtain a server stream appropriate for write operations on the
+ * cluster.
+ *
+ * Returns a new stream (that must be freed) or NULL and sets an error via
+ * `error`.
+ *
+ * @note The returned stream must be released via `mongoc_server_stream_cleanup`
+ *
+ * @note May add nodes and/or update the cluster's topology.
+ */
 mongoc_server_stream_t *
 mongoc_cluster_stream_for_writes (mongoc_cluster_t *cluster,
                                   mongoc_client_session_t *cs,
                                   bson_t *reply,
                                   bson_error_t *error);
 
+/**
+ * @brief Obtain a server stream appropriate for aggregate operations with
+ * writes on the cluster.
+ *
+ * Returns a new stream (that must be freed) or NULL and sets an error via
+ * `error`.
+ *
+ * @note The returned stream must be released via
+ * `mongoc_server_stream_cleanup`.
+ *
+ * @note May add nodes and/or update the cluster's topology.
+ */
+mongoc_server_stream_t *
+mongoc_cluster_stream_for_aggr_with_write (
+   mongoc_cluster_t *cluster,
+   const mongoc_read_prefs_t *read_prefs,
+   mongoc_client_session_t *cs,
+   bson_t *reply,
+   bson_error_t *error);
+
+/**
+ * @brief Obtain a server stream associated with the cluster node associated
+ * with the given server ID.
+ *
+ * Returns a new server stream (that must be freed) or NULL and sets `error`.
+ *
+ * @param server_id The ID of a server in the cluster topology.
+ * @param reconnect_ok If `true`, the server exists in the topology but is not
+ * connected, then attempt to reconnect with the server. If `false`, then only
+ * create a stream if the server is connected and ready.
+ *
+ * @note The returned stream must be released via `mongoc_server_stream_cleanup`
+ *
+ * @note May update the cluster's topology.
+ */
 mongoc_server_stream_t *
 mongoc_cluster_stream_for_server (mongoc_cluster_t *cluster,
                                   uint32_t server_id,
@@ -171,10 +225,9 @@ int
 _mongoc_cluster_get_conversation_id (const bson_t *reply);
 
 mongoc_server_stream_t *
-_mongoc_cluster_create_server_stream (mongoc_topology_t *topology,
-                                      uint32_t server_id,
-                                      mongoc_stream_t *stream,
-                                      bson_error_t *error /* OUT */);
+_mongoc_cluster_create_server_stream (const mongoc_topology_description_t *td,
+                                      const mongoc_server_description_t *sd,
+                                      mongoc_stream_t *stream);
 
 bool
 _mongoc_cluster_get_auth_cmd_x509 (const mongoc_uri_t *uri,
@@ -182,6 +235,16 @@ _mongoc_cluster_get_auth_cmd_x509 (const mongoc_uri_t *uri,
                                    bson_t *cmd /* OUT */,
                                    bson_error_t *error /* OUT */);
 
+/* Returns true if a versioned server API has been selected, otherwise returns
+ * false. */
+bool
+mongoc_cluster_uses_server_api (const mongoc_cluster_t *cluster);
+
+/* Returns true if load balancing mode has been selected, otherwise returns
+ * false. */
+bool
+mongoc_cluster_uses_loadbalanced (const mongoc_cluster_t *cluster);
+
 #ifdef MONGOC_ENABLE_CRYPTO
 void
 _mongoc_cluster_init_scram (const mongoc_cluster_t *cluster,
@@ -195,6 +258,37 @@ _mongoc_cluster_get_auth_cmd_scram (mongoc_crypto_hash_algorithm_t algo,
                                     bson_error_t *error /* OUT */);
 #endif /* MONGOC_ENABLE_CRYPTO */
 
+char *
+_mongoc_rpc_compress (struct _mongoc_cluster_t *cluster,
+                      int32_t compressor_id,
+                      mongoc_rpc_t *rpc_le,
+                      bson_error_t *error);
+bool
+_mongoc_rpc_decompress (mongoc_rpc_t *rpc_le, uint8_t *buf, size_t buflen);
+
+bool
+mcd_rpc_message_compress (mcd_rpc_message *rpc,
+                          int32_t compressor_id,
+                          int32_t compression_level,
+                          void **compressed_data,
+                          size_t *compressed_data_len,
+                          bson_error_t *error);
+
+bool
+mcd_rpc_message_decompress (mcd_rpc_message *rpc,
+                            void **data,
+                            size_t *data_len);
+
+bool
+_mongoc_rpc_decompress_if_necessary (mongoc_rpc_t *rpc,
+                                     mongoc_buffer_t *buffer /* IN/OUT */,
+                                     bson_error_t *error /* OUT */);
+
+bool
+mcd_rpc_message_decompress_if_necessary (mcd_rpc_message *rpc,
+                                         void **data,
+                                         size_t *data_len);
+
 BSON_END_DECLS
 
 
