@@ -14,21 +14,21 @@
  * limitations under the License.
  */
 
-#include "mongoc-prelude.h"
-
 #ifndef MONGOC_RPC_PRIVATE_H
 #define MONGOC_RPC_PRIVATE_H
 
-#include <bson/bson.h>
-#include <stddef.h>
+#include "mongoc-prelude.h"
 
 #include "mongoc-array-private.h"
-#include "mongoc-cmd-private.h"
 #include "mongoc-iovec.h"
-#include "mongoc-write-concern.h"
-#include "mongoc-flags.h"
-/* forward declaration */
-struct _mongoc_cluster_t;
+
+#include "mcd-rpc.h"
+
+#include <bson/bson.h>
+
+#include <stdbool.h>
+#include <stddef.h>
+#include <stdint.h>
 
 BSON_BEGIN_DECLS
 
@@ -74,6 +74,7 @@ typedef struct _mongoc_rpc_section_t {
    const uint8_t *_name;        \
    int32_t _name##_len;
 #define BSON_OPTIONAL(_check, _code) _code
+#define CHECKSUM_FIELD(_name) uint32_t _name;
 
 
 #pragma pack(1)
@@ -104,6 +105,7 @@ typedef union {
    mongoc_rpc_reply_header_t reply_header;
    mongoc_rpc_update_t update;
    mongoc_rpc_compressed_t compressed;
+   int32_t _init; // Used only for initialization.
 } mongoc_rpc_t;
 
 
@@ -128,11 +130,14 @@ BSON_STATIC_ASSERT2 (sizeof_reply_header,
 #undef SECTION_ARRAY_FIELD
 #undef BSON_OPTIONAL
 #undef RAW_BUFFER_FIELD
+#undef CHECKSUM_FIELD
 
 
 void
 _mongoc_rpc_gather (mongoc_rpc_t *rpc, mongoc_array_t *array);
 void
+_mongoc_rpc_op_egress_inc (const mongoc_rpc_t *rpc);
+void
 _mongoc_rpc_swab_to_le (mongoc_rpc_t *rpc);
 void
 _mongoc_rpc_swab_from_le (mongoc_rpc_t *rpc);
@@ -145,22 +150,26 @@ _mongoc_rpc_scatter_reply_header_only (mongoc_rpc_t *rpc,
                                        const uint8_t *buf,
                                        size_t buflen);
 
+bool
+mcd_rpc_message_get_body (const mcd_rpc_message *rpc, bson_t *reply);
+
 bool
 _mongoc_rpc_get_first_document (mongoc_rpc_t *rpc, bson_t *reply);
 bool
 _mongoc_rpc_reply_get_first (mongoc_rpc_reply_t *reply, bson_t *bson);
 bool
 _mongoc_rpc_reply_get_first_msg (mongoc_rpc_msg_t *reply, bson_t *bson);
-
-void
-_mongoc_rpc_prep_command (mongoc_rpc_t *rpc,
-                          const char *cmd_ns,
-                          mongoc_cmd_t *cmd);
 bool
 _mongoc_rpc_check_ok (mongoc_rpc_t *rpc,
                       int32_t error_api_version,
                       bson_error_t *error /* OUT */,
                       bson_t *error_doc /* OUT */);
+
+bool
+mcd_rpc_message_check_ok (mcd_rpc_message *rpc,
+                          int32_t error_api_version,
+                          bson_error_t *error /* OUT */,
+                          bson_t *error_doc /* OUT */);
 bool
 _mongoc_cmd_check_ok (const bson_t *doc,
                       int32_t error_api_version,
@@ -171,19 +180,10 @@ _mongoc_cmd_check_ok_no_wce (const bson_t *doc,
                              int32_t error_api_version,
                              bson_error_t *error);
 
-bool
-_mongoc_rpc_decompress (mongoc_rpc_t *rpc_le, uint8_t *buf, size_t buflen);
-
-char *
-_mongoc_rpc_compress (struct _mongoc_cluster_t *cluster,
-                      int32_t compressor_id,
-                      mongoc_rpc_t *rpc_le,
-                      bson_error_t *error);
-
-bool
-_mongoc_rpc_decompress_if_necessary (mongoc_rpc_t *rpc,
-                                     mongoc_buffer_t *buffer,
-                                     bson_error_t *error);
+void
+mcd_rpc_message_egress (const mcd_rpc_message *rpc);
+void
+mcd_rpc_message_ingress (const mcd_rpc_message *rpc);
 
 BSON_END_DECLS
 
