@@ -140,7 +140,9 @@ static void
 start_thread (json_test_worker_thread_t *wt)
 {
    wt->shutdown_requested = false;
-   mcommon_thread_create (&wt->thread, json_test_worker_thread_run, wt);
+   int ret =
+      mcommon_thread_create (&wt->thread, json_test_worker_thread_run, wt);
+   ASSERT_CMPINT (0, ==, ret);
 }
 
 static void
