@@ -19,6 +19,7 @@
 
 #include <bson/bson.h>
 #include <mongoc/mongoc-buffer-private.h>
+#include <mongoc/mcd-rpc.h>
 
 #include "mongoc/mongoc.h"
 
@@ -30,8 +31,8 @@ struct _mock_server_t; /* forward declaration */
 typedef struct _request_t {
    uint8_t *data;
    size_t data_len;
-   mongoc_rpc_t request_rpc;
-   mongoc_opcode_t opcode; /* copied from rpc for convenience */
+   mcd_rpc_message *rpc;
+   int32_t opcode; /* copied from rpc for convenience */
    struct _mock_server_t *server;
    mongoc_stream_t *client;
    uint16_t client_port;
@@ -52,52 +53,21 @@ request_new (const mongoc_buffer_t *buffer,
              sync_queue_t *replies);
 
 const bson_t *
-request_get_doc (const request_t *request, int n);
+request_get_doc (const request_t *request, size_t n);
 
-bool
-request_matches_flags (const request_t *request, mongoc_query_flags_t flags);
+void
+assert_request_matches_flags (const request_t *request, uint32_t flags);
 
 bool
 request_matches_query (const request_t *request,
                        const char *ns,
-                       mongoc_query_flags_t flags,
+                       uint32_t flags,
                        uint32_t skip,
                        int32_t n_return,
                        const char *query_json,
                        const char *fields_json,
                        bool is_command);
 
-bool
-request_matches_insert (const request_t *request,
-                        const char *ns,
-                        mongoc_insert_flags_t flags,
-                        const char *doc_json);
-
-bool
-request_matches_bulk_insert (const request_t *request,
-                             const char *ns,
-                             mongoc_insert_flags_t flags,
-                             int n);
-
-bool
-request_matches_update (const request_t *request,
-                        const char *ns,
-                        mongoc_update_flags_t flags,
-                        const char *selector_json,
-                        const char *update_json);
-
-bool
-request_matches_delete (const request_t *request,
-                        const char *ns,
-                        mongoc_remove_flags_t flags,
-                        const char *selector_json);
-
-bool
-request_matches_getmore (const request_t *request,
-                         const char *ns,
-                         int32_t n_return,
-                         int64_t cursor_id);
-
 bool
 request_matches_kill_cursors (const request_t *request, int64_t cursor_id);
 
