@@ -35,6 +35,8 @@ get_generation (mongoc_client_t *client, mongoc_cursor_t *cursor)
    uint32_t generation;
    mongoc_server_description_t const *sd;
    bson_error_t error;
+   ASSERT (client);
+
    mc_shared_tpld td = mc_tpld_take_ref (client->topology);
 
    server_id = mongoc_cursor_get_hint (cursor);
@@ -57,6 +59,8 @@ get_connection_count (mongoc_client_t *client)
    bool res;
    int conns;
 
+   ASSERT (client);
+
    BSON_APPEND_INT32 (&cmd, "serverStatus", 1);
    res = mongoc_client_command_simple (
       client, "admin", &cmd, NULL, &reply, &error);
@@ -450,14 +454,14 @@ static void
 _request_error (request_t *request, exhaust_error_type_t error_type)
 {
    if (error_type == NETWORK_ERROR) {
-      mock_server_resets (request);
+      reply_to_request_with_reset (request);
    } else {
-      mock_server_replies (request,
-                           MONGOC_REPLY_QUERY_FAILURE,
-                           123,
-                           0,
-                           0,
-                           "{'$err': 'uh oh', 'code': 4321}");
+      reply_to_request (request,
+                        MONGOC_REPLY_QUERY_FAILURE,
+                        123,
+                        0,
+                        0,
+                        "{'$err': 'uh oh', 'code': 4321}");
    }
 }
 
@@ -469,6 +473,8 @@ _check_error (mongoc_client_t *client,
    uint32_t server_id;
    bson_error_t error;
 
+   ASSERT (client);
+
    server_id = mongoc_cursor_get_hint (cursor);
    ASSERT (server_id);
    ASSERT (mongoc_cursor_error (cursor, &error));
@@ -544,7 +550,7 @@ _mock_test_exhaust (bool pooled,
 
    if (error_when == SECOND_BATCH) {
       /* initial query succeeds, gets a doc and cursor id of 123 */
-      mock_server_replies (request, MONGOC_REPLY_NONE, 123, 1, 1, "{'a': 1}");
+      reply_to_request (request, MONGOC_REPLY_NONE, 123, 1, 1, "{'a': 1}");
       ASSERT (future_get_bool (future));
       assert_match_bson (doc, tmp_bson ("{'a': 1}"), false);
       ASSERT_CMPINT64 ((int64_t) 123, ==, mongoc_cursor_get_id (cursor));
