@@ -925,7 +925,7 @@ test_mongoc_handshake_race_condition (void)
 
       for (j = 0; j < 4; ++j) {
          BSON_ASSERT (!mcommon_thread_create (
-            &threads[j], &handshake_append_worker, NULL));
+            &threads[j], &handshake_append_worker, NULL /* args */));
       }
       for (j = 0; j < 4; ++j) {
          mcommon_thread_join (threads[j]);
