@@ -45,12 +45,12 @@ _test_op_msg (const mongoc_uri_t *uri,
 
    future = future_cursor_next (cursor, &doc);
    request = mock_server_receives_msg (server, 0, tmp_bson (expected_find));
-   mock_server_replies_simple (request,
-                               "{'ok': 1,"
-                               " 'cursor': {"
-                               "    'id': 0,"
-                               "    'ns': 'db.collection',"
-                               "    'firstBatch': [{'a': 1}]}}");
+   reply_to_request_simple (request,
+                            "{'ok': 1,"
+                            " 'cursor': {"
+                            "    'id': 0,"
+                            "    'ns': 'db.collection',"
+                            "    'firstBatch': [{'a': 1}]}}");
 
    /* mongoc_cursor_next returned true */
    BSON_ASSERT (future_get_bool (future));
@@ -97,12 +97,12 @@ _test_command (const mongoc_uri_t *uri,
    request = mock_server_receives_msg (
       server, MONGOC_MSG_NONE, tmp_bson (expected_cmd));
 
-   mock_server_replies (request,
-                        MONGOC_REPLY_NONE, /* flags */
-                        0,                 /* cursorId */
-                        0,                 /* startingFrom */
-                        1,                 /* numberReturned */
-                        "{'ok': 1}");
+   reply_to_request (request,
+                     MONGOC_REPLY_NONE, /* flags */
+                     0,                 /* cursorId */
+                     0,                 /* startingFrom */
+                     1,                 /* numberReturned */
+                     "{'ok': 1}");
 
    /* mongoc_cursor_next returned true */
    BSON_ASSERT (future_get_bool (future));
@@ -137,12 +137,12 @@ _test_command_simple (const mongoc_uri_t *uri,
    request = mock_server_receives_msg (
       server, MONGOC_MSG_NONE, tmp_bson (expected_cmd));
 
-   mock_server_replies (request,
-                        MONGOC_REPLY_NONE, /* flags */
-                        0,                 /* cursorId */
-                        0,                 /* startingFrom */
-                        1,                 /* numberReturned */
-                        "{'ok': 1}");
+   reply_to_request (request,
+                     MONGOC_REPLY_NONE, /* flags */
+                     0,                 /* cursorId */
+                     0,                 /* startingFrom */
+                     1,                 /* numberReturned */
+                     "{'ok': 1}");
 
    /* mongoc_cursor_next returned true */
    BSON_ASSERT (future_get_bool (future));
@@ -216,8 +216,7 @@ _run_server (read_pref_test_type_t test_type)
                               mock_server_get_host_and_port (server));
       break;
    default:
-      fprintf (stderr, "Invalid test_type: : %d\n", test_type);
-      abort ();
+      test_error ("Invalid test_type: %d", (int) test_type);
    }
 
    return server;
@@ -612,14 +611,13 @@ test_read_prefs_mongos_max_staleness (void)
       "                     'maxStalenessSeconds': 120}}",
       "{}");
 
-   mock_server_replies_to_find (request,
-                                MONGOC_QUERY_EXHAUST |
-                                   MONGOC_QUERY_SECONDARY_OK,
-                                0,
-                                1,
-                                "test.test",
-                                "{}",
-                                false);
+   reply_to_find_request (request,
+                          MONGOC_QUERY_EXHAUST | MONGOC_QUERY_SECONDARY_OK,
+                          0,
+                          1,
+                          "test.test",
+                          "{}",
+                          false);
 
    /* mongoc_cursor_next returned true */
    BSON_ASSERT (future_get_bool (future));
@@ -674,14 +672,13 @@ test_read_prefs_mongos_hedged_reads (void)
       "                     'hedge': {'enabled': true}}}",
       "{}");
 
-   mock_server_replies_to_find (request,
-                                MONGOC_QUERY_EXHAUST |
-                                   MONGOC_QUERY_SECONDARY_OK,
-                                0,
-                                1,
-                                "test.test",
-                                "{}",
-                                false);
+   reply_to_find_request (request,
+                          MONGOC_QUERY_EXHAUST | MONGOC_QUERY_SECONDARY_OK,
+                          0,
+                          1,
+                          "test.test",
+                          "{}",
+                          false);
 
    /* mongoc_cursor_next returned true */
    BSON_ASSERT (future_get_bool (future));
@@ -733,8 +730,13 @@ test_mongos_read_concern (void)
                 " '$readPreference': {'mode': 'secondary'},"
                 " 'readConcern': {'level': 'foo'}}"));
 
-   mock_server_replies_to_find (
-      request, MONGOC_QUERY_NONE, 0, 1, "db.collection", "{}", true);
+   reply_to_op_msg_request (request,
+                            MONGOC_MSG_NONE,
+                            tmp_bson ("{'ok': 1,"
+                                      " 'cursor': {"
+                                      "    'id': {'$numberLong': '0'},"
+                                      "    'ns': 'db.collection',"
+                                      "    'firstBatch': [{}]}}"));
 
    /* mongoc_cursor_next returned true */
    BSON_ASSERT (future_get_bool (future));
@@ -773,7 +775,7 @@ _test_op_msg_direct_connection (bool is_mongos,
    int i;
 
    if (is_mongos) {
-      server = mock_mongos_new (WIRE_VERSION_OP_MSG);
+      server = mock_mongos_new (WIRE_VERSION_MAX);
    } else {
       char *hello = bson_strdup_printf ("{'ok': 1.0,"
                                         " 'isWritablePrimary': true,"
@@ -782,7 +784,7 @@ _test_op_msg_direct_connection (bool is_mongos,
                                         " 'minWireVersion': %d,"
                                         " 'maxWireVersion': %d}",
                                         WIRE_VERSION_MIN,
-                                        WIRE_VERSION_OP_MSG);
+                                        WIRE_VERSION_MAX);
       server = mock_server_new ();
       mock_server_auto_hello (server, hello);
       bson_free (hello);
@@ -811,7 +813,7 @@ _test_op_msg_direct_connection (bool is_mongos,
               "    'ns': 'db.collection',"
               "    'firstBatch': [{'a': 1}]}}";
 
-      mock_server_replies_simple (request, reply);
+      reply_to_request_simple (request, reply);
       BSON_ASSERT (future_get_bool (future));
       future_destroy (future);
       request_destroy (request);
