@@ -70,6 +70,8 @@ _set_failpoint (mongoc_client_t *client)
                 " 'mode': {'times': 1},"
                 " 'data': {'errorCode': 10107, 'failCommands': ['count']}}");
 
+   ASSERT (client);
+
    ASSERT_OR_PRINT (
       mongoc_client_command_simple (client, "admin", cmd, NULL, NULL, &error),
       error);
@@ -93,7 +95,7 @@ test_cmd_helpers (void *ctx)
    BSON_UNUSED (ctx);
 
    uri = test_framework_get_uri ();
-   mongoc_uri_set_option_as_bool (uri, "retryReads", true);
+   mongoc_uri_set_option_as_bool (uri, MONGOC_URI_RETRYREADS, true);
 
    client = test_framework_client_new_from_uri (uri, NULL);
    mongoc_client_set_error_api (client, MONGOC_ERROR_API_VERSION_2);
@@ -288,7 +290,7 @@ test_all_spec_tests (TestSuite *suite)
 {
    install_json_test_suite_with_check (suite,
                                        JSON_DIR,
-                                       "retryable_reads",
+                                       "retryable_reads/legacy",
                                        test_retryable_reads_cb,
                                        TestSuite_CheckLive,
                                        test_framework_skip_if_no_failpoint,
