@@ -19,45 +19,6 @@
 #include "test-conveniences.h"
 #include "TestSuite.h"
 
-uint8_t *
-hex_to_bin (const char *hex, uint32_t *len)
-{
-   int i;
-   int hex_len;
-   uint8_t *out;
-
-   hex_len = strlen (hex);
-   if (hex_len % 2 != 0) {
-      return NULL;
-   }
-
-   *len = hex_len / 2;
-   out = bson_malloc0 (*len);
-
-   for (i = 0; i < hex_len; i += 2) {
-      uint32_t hex_char;
-
-      if (1 != sscanf (hex + i, "%2x", &hex_char)) {
-         bson_free (out);
-         return NULL;
-      }
-      out[i / 2] = (uint8_t) hex_char;
-   }
-   return out;
-}
-
-char *
-bin_to_hex (const uint8_t *bin, uint32_t len)
-{
-   char *out = bson_malloc0 (2 * len + 1);
-   uint32_t i;
-
-   for (i = 0; i < len; i++) {
-      bson_snprintf (out + (2 * i), 2, "%02x", bin[i]);
-   }
-   return out;
-}
-
 static int
 cmp_key (const void *a, const void *b)
 {
@@ -94,7 +55,7 @@ typedef struct {
    bson_type_t type;
 } bson_string_and_type_t;
 
-/* List of aliases: https://docs.mongodb.com/manual/reference/bson-types/ */
+/* List of aliases: https://www.mongodb.com/docs/manual/reference/bson-types/ */
 bson_string_and_type_t bson_type_map[] = {
    {"double", BSON_TYPE_DOUBLE},
    {"string", BSON_TYPE_UTF8},
@@ -185,9 +146,29 @@ is_unsupported_event_type (const char *event_type)
    char **iter;
 
    for (iter = unsupported_event_types; *iter != NULL; iter++) {
-      if (0 == strcmp (event_type, *iter)) {
+      if (0 == bson_strcasecmp (event_type, *iter)) {
          return true;
       }
    }
    return false;
 }
+
+int64_t
+usecs_since_epoch (void)
+{
+   struct timeval tv;
+   BSON_ASSERT (bson_gettimeofday (&tv) == 0);
+
+   BSON_ASSERT (bson_in_range_signed (int64_t, tv.tv_sec));
+   BSON_ASSERT (bson_in_range_signed (int64_t, tv.tv_usec));
+
+   const int64_t secs = (int64_t) tv.tv_sec;
+   const int64_t usecs = (int64_t) tv.tv_usec;
+
+   const int64_t factor = 1000000;
+
+   BSON_ASSERT (INT64_MAX / factor >= secs);
+   BSON_ASSERT (INT64_MAX - (factor * secs) >= usecs);
+
+   return secs * factor + usecs;
+}
